@@ -6,7 +6,7 @@
 ;; URL: http://github.com/s-kostyaev/ellama
 ;; Keywords: help local tools
 ;; Package-Requires: ((emacs "28.1") (llm "0.6.0") (spinner "1.7.4"))
-;; Version: 0.8.13
+;; Version: 0.9.11
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 ;; Created: 8th Oct 2023
 
@@ -35,10 +35,13 @@
 
 ;;; Code:
 
+(require 'eieio)
 (require 'json)
 (require 'llm)
 (require 'spinner)
 (require 'info)
+(require 'shr)
+(require 'eww)
 (eval-when-compile (require 'rx))
 
 (defgroup ellama nil
@@ -109,6 +112,7 @@
     ;; summarize
     (define-key map (kbd "s s") 'ellama-summarize)
     (define-key map (kbd "s w") 'ellama-summarize-webpage)
+    (define-key map (kbd "s c") 'ellama-summarize-killring)
     ;; session
     (define-key map (kbd "s l") 'ellama-load-session)
     (define-key map (kbd "s r") 'ellama-session-rename)
@@ -319,6 +323,11 @@ Too low value can break generated code by splitting long comment lines."
   :group 'ellama
   :type 'boolean)
 
+(defcustom ellama-show-quotes nil
+  "Show quotes in chat context."
+  :group 'ellama
+  :type 'boolean)
+
 (define-minor-mode ellama-session-mode
   "Minor mode for ellama session buffers."
   :interactive nil
@@ -345,7 +354,7 @@ Too low value can break generated code by splitting long comment lines."
 
 (defconst ellama--code-prefix
   (rx (minimal-match
-       (zero-or-more anything) (literal "```") (zero-or-more anything) line-end)))
+       (zero-or-more anything) (literal "```") (zero-or-more anything) (+ (or "\n" "\r")))))
 
 (defconst ellama--code-suffix
   (rx (minimal-match
@@ -365,43 +374,65 @@ Too low value can break generated code by splitting long comment lines."
       (fill-region (point-min) (point-max) nil t t))
     (buffer-substring-no-properties (point-min) (point-max))))
 
+(defun ellama--replace-first-begin-src (text)
+  "Replace first begin src in TEXT."
+  (if (not (string-match-p (rx (literal "#+BEGIN_SRC")) text))
+      (replace-regexp-in-string "^[[:space:]]*```\\(\\(.\\|\n\\)*\\)" "#+BEGIN_SRC\\1" text)
+    text))
+
+(defun ellama--replace-bad-code-blocks (text)
+  "Replace code src blocks in TEXT."
+  (with-temp-buffer
+    (insert text)
+    (goto-char (point-min))
+    ;; skip good code blocks
+    (while (re-search-forward "#\\+BEGIN_SRC\\(.\\|\n\\)*?#\\+END_SRC" nil t))
+    (while (re-search-forward "#\\+END_SRC\\(\\(.\\|\n\\)*?\\)#\\+END_SRC" nil t)
+      (replace-match "#+BEGIN_SRC\\1#+END_SRC"))
+    (buffer-substring-no-properties (point-min) (point-max))))
+
 (defun ellama--translate-markdown-to-org-filter (text)
   "Filter to translate code blocks from markdown syntax to org syntax in TEXT.
 This filter contains only subset of markdown syntax to be good enough."
-  (thread-last text
-       ;; code blocks
-       (replace-regexp-in-string "^[[:space:]]*```\\(.+\\)$" "#+BEGIN_SRC \\1")
-       (replace-regexp-in-string "^<!-- language: \\(.+\\) -->\n```" "#+BEGIN_SRC \\1")
-       (replace-regexp-in-string "^[[:space:]]*```$" "#+END_SRC")
-       ;; lists
-       (replace-regexp-in-string "^\\* " "+ ")
-       ;; bold
-       (replace-regexp-in-string "\\*\\*\\(.+?\\)\\*\\*" "*\\1*")
-       (replace-regexp-in-string "__\\(.+?\\)__" "*\\1*")
-       (replace-regexp-in-string "<b>\\(.+?\\)</b>" "*\\1*")
-       ;; italic
-       ;; (replace-regexp-in-string "_\\(.+?\\)_" "/\\1/") ;; most of the time it breaks code blocks, so disable it
-       (replace-regexp-in-string "<i>\\(.+?\\)</i>" "/\\1/")
-       ;; inline code
-       (replace-regexp-in-string "`\\(.+?\\)`" "~\\1~")
-       ;; underlined
-       (replace-regexp-in-string "<u>\\(.+?\\)</u>" "_\\1_")
-       ;; strikethrough
-       (replace-regexp-in-string "~~\\(.+?\\)~~" "+\\1+")
-       (replace-regexp-in-string "<s>\\(.+?\\)</s>" "+\\1+")
-       ;; headings
-       (replace-regexp-in-string "^# " "* ")
-       (replace-regexp-in-string "^## " "** ")
-       (replace-regexp-in-string "^### " "*** ")
-       (replace-regexp-in-string "^#### " "**** ")
-       (replace-regexp-in-string "^##### " "***** ")
-       (replace-regexp-in-string "^###### " "***** ")
-       ;; badges
-       (replace-regexp-in-string "\\[\\!\\[.*?\\](\\(.*?\\))\\](\\(.*?\\))" "[[\\2][file:\\1]]")
-       ;;links
-       (replace-regexp-in-string "\\[\\(.*?\\)\\](\\(.*?\\))" "[[\\2][\\1]]")
-       ;; filling long lines
-       (ellama--fill-long-lines)))
+  (thread-last
+    text
+    ;; code blocks
+    (replace-regexp-in-string "^[[:space:]]*```\\(.+\\)$" "#+BEGIN_SRC \\1")
+    (ellama--replace-first-begin-src)
+    (replace-regexp-in-string "^<!-- language: \\(.+\\) -->\n```" "#+BEGIN_SRC \\1")
+    (replace-regexp-in-string "^[[:space:]]*```$" "#+END_SRC")
+    (replace-regexp-in-string "^[[:space:]]*```" "#+END_SRC\n")
+    (replace-regexp-in-string "```" "\n#+END_SRC\n")
+    (ellama--replace-bad-code-blocks)
+    ;; lists
+    (replace-regexp-in-string "^\\* " "+ ")
+    ;; bold
+    (replace-regexp-in-string "\\*\\*\\(.+?\\)\\*\\*" "*\\1*")
+    (replace-regexp-in-string "__\\(.+?\\)__" "*\\1*")
+    (replace-regexp-in-string "<b>\\(.+?\\)</b>" "*\\1*")
+    ;; italic
+    ;; (replace-regexp-in-string "_\\(.+?\\)_" "/\\1/") ;; most of the time it breaks code blocks, so disable it
+    (replace-regexp-in-string "<i>\\(.+?\\)</i>" "/\\1/")
+    ;; inline code
+    (replace-regexp-in-string "`\\(.+?\\)`" "~\\1~")
+    ;; underlined
+    (replace-regexp-in-string "<u>\\(.+?\\)</u>" "_\\1_")
+    ;; strikethrough
+    (replace-regexp-in-string "~~\\(.+?\\)~~" "+\\1+")
+    (replace-regexp-in-string "<s>\\(.+?\\)</s>" "+\\1+")
+    ;; headings
+    (replace-regexp-in-string "^# " "* ")
+    (replace-regexp-in-string "^## " "** ")
+    (replace-regexp-in-string "^### " "*** ")
+    (replace-regexp-in-string "^#### " "**** ")
+    (replace-regexp-in-string "^##### " "***** ")
+    (replace-regexp-in-string "^###### " "***** ")
+    ;; badges
+    (replace-regexp-in-string "\\[\\!\\[.*?\\](\\(.*?\\))\\](\\(.*?\\))" "[[\\2][file:\\1]]")
+    ;;links
+    (replace-regexp-in-string "\\[\\(.*?\\)\\](\\(.*?\\))" "[[\\2][\\1]]")
+    ;; filling long lines
+    (ellama--fill-long-lines)))
 
 (defcustom ellama-enable-keymap t
   "Enable or disable Ellama keymap."
@@ -427,6 +458,11 @@ This filter contains only subset of markdown syntax to be good enough."
   :group 'ellama
   :type '(sexp :validate 'cl-struct-p))
 
+(defcustom ellama-always-show-chain-steps nil
+  "Always show ellama chain buffers."
+  :type 'boolean
+  :group 'ellama)
+
 (defvar-local ellama--current-session nil)
 
 (defvar ellama--current-session-id nil)
@@ -531,8 +567,16 @@ If EPHEMERAL non nil new session will not be associated with any file."
 		      (file-name-concat
 		       ellama-sessions-directory
 		       (concat id "." (ellama-get-session-file-extension)))))
+ 	 (previous-session
+	  (when ellama--current-session-id
+	    (with-current-buffer
+		(ellama-get-session-buffer ellama--current-session-id)
+	      ellama--current-session)))
 	 (session (make-ellama-session
-		   :id id :provider provider :file file-name :context ellama--new-session-context))
+		   :id id :provider provider :file file-name
+		   :context (if previous-session
+				(ellama-session-context previous-session)
+			      ellama--new-session-context)))
 	 (buffer (if file-name
 		     (progn
 		       (make-directory ellama-sessions-directory t)
@@ -729,48 +773,368 @@ If EPHEMERAL non nil new session will not be associated with any file."
     (remhash id ellama--active-sessions)
     (puthash new-id buffer ellama--active-sessions)))
 
+;; Context elements
+
+(defclass ellama-context-element () ()
+  "A structure for holding information about a context element.")
+
+(cl-defgeneric ellama-context-element-add (element)
+  "Add the ELEMENT to the Ellama context.")
+
+(cl-defgeneric ellama-context-element-extract (element)
+  "Extract the content of the context ELEMENT.")
+
+(cl-defgeneric ellama-context-element-format (element mode)
+  "Format the context ELEMENT for the major MODE.")
+
+(cl-defmethod ellama-context-element-add ((element ellama-context-element))
+  "Add the ELEMENT to the Ellama context."
+  (if-let* ((id ellama--current-session-id)
+	    (session (with-current-buffer (ellama-get-session-buffer id)
+		       ellama--current-session)))
+      (push element (ellama-session-context session))
+    (push element ellama--new-session-context)))
+
+;; Buffer context element
+
+(defclass ellama-context-element-buffer (ellama-context-element)
+  ((name :initarg :name :type string))
+  "A structure for holding information about a context element.")
+
+(cl-defmethod ellama-context-element-extract
+  ((element ellama-context-element-buffer))
+  "Extract the content of the context ELEMENT."
+  (with-slots (name) element
+    (with-current-buffer name
+      (buffer-substring-no-properties (point-min) (point-max)))))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-buffer) (mode (eql 'markdown-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name) element
+    (format "```emacs-lisp\n(display-buffer \"%s\")\n```\n" name)))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-buffer) (mode (eql 'org-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name) element
+    (format "[[elisp:(display-buffer \"%s\")][%s]]" name name)))
+
+;; File context element
+
+(defclass ellama-context-element-file (ellama-context-element)
+  ((name :initarg :name :type string))
+  "A structure for holding information about a context element.")
+
+(cl-defmethod ellama-context-element-extract
+  ((element ellama-context-element-file))
+  "Extract the content of the context ELEMENT."
+  (with-slots (name) element
+    (with-temp-buffer
+      (insert-file-contents name)
+      (buffer-substring-no-properties (point-min) (point-max)))))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-file) (mode (eql 'markdown-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name) element
+    (format "[%s](<%s>)" name name)))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-file) (mode (eql 'org-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name) element
+    (format "[[file:%s][%s]]" name name)))
+
+;; Info node context element
+
+(defclass ellama-context-element-info-node (ellama-context-element)
+  ((name :initarg :name :type string))
+  "A structure for holding information about a context element.")
+
+(cl-defmethod ellama-context-element-extract
+  ((element ellama-context-element-info-node))
+  "Extract the content of the context ELEMENT."
+  (with-slots (name) element
+    (with-temp-buffer
+      (info name (current-buffer))
+      (buffer-substring-no-properties (point-min) (point-max)))))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-info-node) (mode (eql 'markdown-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name) element
+    (format "```emacs-lisp\n(info \"%s\")\n```\n" name)))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-info-node) (mode (eql 'org-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name) element
+    (format "[[%s][%s]]"
+	    (replace-regexp-in-string
+	     "(\\(.?*\\)) \\(.*\\)" "info:\\1#\\2" name)
+	    (if (and ellama-chat-translation-enabled
+		     (not ellama--current-session))
+		(ellama--translate-string name)
+	      name))))
+
+;; Text context element
+
+(defclass ellama-context-element-text (ellama-context-element)
+  ((content :initarg :content :type string))
+  "A structure for holding information about a context element.")
+
+(cl-defmethod ellama-context-element-extract
+  ((element ellama-context-element-text))
+  "Extract the content of the context ELEMENT."
+  (oref element content))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-text) (mode (eql 'markdown-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (oref element content))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-text) (mode (eql 'org-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (oref element content))
+
+;; Webpage quote context elements
+
+(defclass ellama-context-element-webpage-quote (ellama-context-element)
+  ((name :initarg :name :type string)
+   (url :initarg :url :type string)
+   (content :initarg :content :type string))
+  "A structure for holding information about a context element.")
+
+(cl-defmethod ellama-context-element-extract
+  ((element ellama-context-element-webpage-quote))
+  "Extract the content of the context ELEMENT."
+  (oref element content))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-webpage-quote) (mode (eql 'markdown-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name url content) element
+    (if ellama-show-quotes
+	(format "[%s](%s):\n%s\n\n"
+		name url
+		(ellama--md-quote content))
+      (format "[%s](%s)" name url))))
+
+(defun ellama--md-quote (content)
+  "Return quoted CONTENT for markdown."
+  (with-temp-buffer
+    (insert (propertize content 'hard t))
+    (let ((fill-prefix "> ")
+	  (fill-column ellama-long-lines-length)
+	  (use-hard-newlines t)
+	  (comment-start ">")
+	  (comment-empty-lines t))
+      (comment-region (point-min) (point-max) ">")
+      (fill-region (point-min) (point-max) nil t t))
+    (buffer-substring-no-properties (point-min) (point-max))))
+
+(defun ellama--org-quote (content)
+  "Return transformed CONTENT for org quotes."
+  (replace-regexp-in-string "^*" " *" content))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-webpage-quote) (mode (eql 'org-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name url content) element
+    (if ellama-show-quotes
+	(format "[[%s][%s]]:\n#+BEGIN_QUOTE\n%s\n#+END_QUOTE\n"
+		url name (ellama--org-quote content))
+      (format "[[%s][%s]]" url name))))
+
+;; Info node quote context elements
+
+(defclass ellama-context-element-info-node-quote (ellama-context-element)
+  ((name :initarg :name :type string)
+   (content :initarg :content :type string))
+  "A structure for holding information about a context element.")
+
+(cl-defmethod ellama-context-element-extract
+  ((element ellama-context-element-info-node-quote))
+  "Extract the content of the context ELEMENT."
+  (oref element content))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-info-node-quote) (mode (eql 'markdown-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name content) element
+    (if ellama-show-quotes
+	(format "```emacs-lisp\n(info \"%s\")\n```\n%s\n\n"
+		name
+		(ellama--md-quote content))
+      (format "```emacs-lisp\n(info \"%s\")\n```\n" name))))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-info-node-quote) (mode (eql 'org-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (name content) element
+    (if ellama-show-quotes
+	(format "[[%s][%s]]:\n#+BEGIN_QUOTE\n%s\n#+END_QUOTE\n"
+		(replace-regexp-in-string
+		 "(\\(.?*\\)) \\(.*\\)" "info:\\1#\\2" name)
+		(if (and ellama-chat-translation-enabled
+			 (not ellama--current-session))
+		    (ellama--translate-string name)
+		  name)
+		(ellama--org-quote content))
+      (format "[[%s][%s]]"
+	      (replace-regexp-in-string
+	       "(\\(.?*\\)) \\(.*\\)" "info:\\1#\\2" name)
+	      (if (and ellama-chat-translation-enabled
+		       (not ellama--current-session))
+		  (ellama--translate-string name)
+		name)))))
+
+;; File quote context elements
+
+(defclass ellama-context-element-file-quote (ellama-context-element)
+  ((path :initarg :path :type string)
+   (content :initarg :content :type string))
+  "A structure for holding information about a context element.")
+
+(cl-defmethod ellama-context-element-extract
+  ((element ellama-context-element-file-quote))
+  "Extract the content of the context ELEMENT."
+  (oref element content))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-file-quote) (mode (eql 'markdown-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (path content) element
+    (if ellama-show-quotes
+	(format "[%s](%s):\n%s\n\n"
+		path path
+		(ellama--md-quote content))
+      (format "[%s](%s)" path path))))
+
+(cl-defmethod ellama-context-element-format
+  ((element ellama-context-element-file-quote) (mode (eql 'org-mode)))
+  "Format the context ELEMENT for the major MODE."
+  (ignore mode)
+  (with-slots (path content) element
+    (if ellama-show-quotes
+	(format "[[%s][%s]]:\n#+BEGIN_QUOTE\n%s\n#+END_QUOTE\n"
+		path path (ellama--org-quote content))
+      (format "[[%s][%s]]" path path))))
+
+
 ;;;###autoload
 (defun ellama-context-add-file ()
   "Add file to context."
   (interactive)
-  (if-let* ((id ellama--current-session-id)
-	    (session (with-current-buffer (ellama-get-session-buffer id)
-		       ellama--current-session))
-	    (file-name (read-file-name "Select file: " nil nil t)))
-      (push (cons 'file file-name) (ellama-session-context session))
-    (push (cons 'file file-name) ellama--new-session-context)))
+  (let* ((file-name (read-file-name "Select file: " nil nil t))
+         (element (ellama-context-element-file :name file-name)))
+    (ellama-context-element-add element)))
+
+(defun ellama-context-add-file-quote-noninteractive (path content)
+  "Add file with PATH quote CONTENT to context."
+  (let ((element (ellama-context-element-file-quote
+		  :path path :content content)))
+    (ellama-context-element-add element)))
+
+;;;###autoload
+(defun ellama-context-add-file-quote ()
+  "Add file quote to context interactively."
+  (interactive)
+  (let ((path (buffer-file-name (current-buffer)))
+	(content (if (region-active-p)
+		     (buffer-substring-no-properties
+		      (region-beginning)
+		      (region-end))
+		   (buffer-substring-no-properties
+		    (point-min)
+		    (point-max)))))
+    (if (not path)
+	(warn "should be called from buffer associated with file")
+      (ellama-context-add-file-quote-noninteractive path content))))
 
 ;;;###autoload
 (defun ellama-context-add-buffer (buf)
   "Add BUF to context."
   (interactive "bSelect buffer: ")
-  (if-let* ((id ellama--current-session-id)
-	    (session (with-current-buffer (ellama-get-session-buffer id)
-		       ellama--current-session)))
-      (push (cons 'buffer buf) (ellama-session-context session))
-    (push (cons 'buffer buf) ellama--new-session-context)))
+  (let ((element (ellama-context-element-buffer :name buf)))
+    (ellama-context-element-add element)))
 
 ;;;###autoload
 (defun ellama-context-add-selection ()
   "Add file to context."
   (interactive)
-  (if-let* ((id ellama--current-session-id)
-	    (session (with-current-buffer (ellama-get-session-buffer id)
-		       ellama--current-session))
-	    ((region-active-p))
-	    (content (buffer-substring-no-properties (region-beginning) (region-end))))
-      (push (cons 'text content) (ellama-session-context session))
-    (push (cons 'text content) ellama--new-session-context)))
+  (if (region-active-p)
+      (let* ((content (buffer-substring-no-properties (region-beginning) (region-end)))
+             (element (ellama-context-element-text :content content)))
+        (ellama-context-element-add element))
+    (warn "No active region")))
 
 ;;;###autoload
 (defun ellama-context-add-info-node (node)
   "Add info NODE to context."
   (interactive (list (Info-copy-current-node-name)))
-  (if-let* ((id ellama--current-session-id)
-	    (session (with-current-buffer (ellama-get-session-buffer id)
-		       ellama--current-session)))
-      (push (cons 'info node) (ellama-session-context session))
-    (push (cons 'info node) ellama--new-session-context)))
+  (let ((element (ellama-context-element-info-node :name node)))
+    (ellama-context-element-add element)))
+
+(defun ellama-context-add-info-node-quote-noninteractive (name content)
+  "Add info node with NAME quote CONTENT to context."
+  (let ((element (ellama-context-element-info-node-quote
+		  :name name :content content)))
+    (ellama-context-element-add element)))
+
+;;;###autoload
+(defun ellama-context-add-info-node-quote ()
+  "Add info node quote to context interactively."
+  (interactive)
+  (let ((name (Info-copy-current-node-name))
+	(content (if (region-active-p)
+		     (buffer-substring-no-properties
+		      (region-beginning)
+		      (region-end))
+		   (buffer-substring-no-properties
+		    (point-min)
+		    (point-max)))))
+    (if (not name)
+	(warn "should be called from `info' buffer")
+      (ellama-context-add-info-node-quote-noninteractive name content))))
+
+(defun ellama-context-add-webpage-quote-noninteractive (name url content)
+  "Add webpage with NAME and URL quote CONTENT to context."
+  (let ((element (ellama-context-element-webpage-quote
+		  :name name :url url :content content)))
+    (ellama-context-element-add element)))
+
+;;;###autoload
+(defun ellama-context-add-webpage-quote-eww ()
+  "Add webpage quote to context interactively from `eww'."
+  (interactive)
+  (if (eq major-mode 'eww-mode)
+      (let* ((name (plist-get eww-data :title))
+	     (url (eww-current-url))
+	     (content (if (region-active-p)
+			  (buffer-substring-no-properties
+			   (region-beginning)
+			   (region-end))
+			(buffer-substring-no-properties
+			 (point-min)
+			 (point-max)))))
+	(ellama-context-add-webpage-quote-noninteractive name url content))
+    (warn "Should be called from `eww'.")))
 
 (defun ellama--translate-string (s)
   "Translate string S to `ellama-language' syncronously."
@@ -782,68 +1146,18 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	    s
 	    ellama-language))))
 
-(defun ellama--org-format-context-element (elt)
-  "Format context ELT for org mode."
-  (pcase (car elt)
-    ('file
-     (format "[[file:%s][%s]]" (cdr elt) (cdr elt)))
-    ('buffer
-     (format "[[elisp:(display-buffer \"%s\")][%s]]" (cdr elt) (cdr elt)))
-    ('text
-     (cdr elt))
-    ('info
-     (format "[[%s][%s]]"
-	     (replace-regexp-in-string
-	      "(\\(.?*\\)) \\(.*\\)" "info:\\1#\\2" (cdr elt))
-	     (if (and ellama-chat-translation-enabled
-		      (not ellama--current-session))
-		 (ellama--translate-string (cdr elt))
-	       (cdr elt))))
-    (_
-     (user-error "Unsupported context element"))))
-
-(defun ellama--md-format-context-element (elt)
-  "Format context ELT for org mode."
-  (pcase (car elt)
-    ('file
-     (format "[%s](<%s>)"
-	     (cdr elt)
-	     (cdr elt)))
-    ('buffer
-     (format "```emacs-lisp\n(display-buffer \"%s\")\n```\n" (cdr elt)))
-    ('text
-     (cdr elt))
-    ('info
-     (format "```emacs-lisp\n(info \"%s\")\n```\n" (cdr elt)))
-    (_
-     (user-error "Unsupported context element"))))
-
-(defun ellama--extract-context-element (elt)
-  "Extract context ELT content."
-  (pcase (car elt)
-    ('text (cdr elt))
-    ('buffer (with-current-buffer (cdr elt)
-	       (buffer-substring-no-properties (point-min) (point-max))))
-    ('file (with-temp-buffer
-	     (find-file-literally (cdr elt))
-	     (buffer-substring-no-properties (point-min) (point-max))))
-    ('info (with-temp-buffer
-	     (info (cdr elt) (current-buffer))
-	     (buffer-substring-no-properties (point-min) (point-max))))
-    (_
-     (user-error "Unsupported context element"))))
-
 (defun ellama--format-context (session)
   "Format SESSION context for chat buffer."
-  (if-let* ((context (ellama-session-context session)))
-      (concat (string-join
-	       (cons "Context:"
-		     (if (derived-mode-p 'org-mode)
-			 (mapcar #'ellama--org-format-context-element context)
-		       (mapcar #'ellama--md-format-context-element context)))
-	       "\n")
-	      "\n\n")
-    ""))
+  (let ((mode (if (derived-mode-p 'org-mode) 'org-mode 'markdown-mode)))
+    (if-let* ((context (ellama-session-context session)))
+        (concat (string-join
+	         (cons "Context:"
+                       (mapcar (lambda (elt)
+                                 (ellama-context-element-format elt mode))
+                               context))
+	         "\n")
+	        "\n\n")
+      "")))
 
 (defun ellama--prompt-with-context (prompt)
   "Add context to PROMPT for sending to llm."
@@ -851,7 +1165,7 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	    (context (ellama-session-context session)))
       (concat (string-join
 	       (cons "Context:"
-		     (mapcar #'ellama--extract-context-element context))
+		     (mapcar #'ellama-context-element-extract context))
 	       "\n")
 	      "\n\n"
 	      prompt)
@@ -964,6 +1278,50 @@ when the request completes (with BUFFER current)."
 				      (setq ellama--current-request nil)
 				      (ellama-request-mode -1)))))))))
 
+(defun ellama-chain (initial-prompt forms)
+  "Call chain of FORMS on INITIAL-PROMPT.
+Each form is a plist that can contain different options:
+
+:provider PROVIDER - use PROVIDER instead of `ellama-provider'.
+
+:transform FUNCTION - use FUNCTION to transform result of previous step to new
+prompt.
+
+:session SESSION - use SESSION in current step.
+
+:chat BOOL - if BOOL use chat buffer, otherwise use temp buffer.  Make sense for
+last step only.
+
+:show BOOL - if BOOL show buffer for this step."
+  (let* ((hd (car forms))
+	 (tl (cdr forms))
+	 (provider (or (plist-get hd :provider) ellama-provider))
+	 (transform (or (plist-get hd :transform) #'identity))
+	 (prompt (apply transform (list initial-prompt)))
+	 (session (plist-get hd :session))
+	 (chat (plist-get hd :chat))
+	 (show (or (plist-get hd :show) ellama-always-show-chain-steps chat))
+	 (buf (if (or (and tl (not chat)) (not session))
+		  (get-buffer-create (make-temp-name
+				      (ellama-generate-name provider real-this-command prompt)))
+		(ellama-get-session-buffer ellama--current-session-id))))
+    (when show
+      (display-buffer buf))
+    (with-current-buffer buf
+      (funcall ellama-major-mode))
+    (if chat
+	(ellama-chat prompt nil :provider provider)
+      (ellama-stream
+       prompt
+       :provider provider
+       :buffer buf
+       :session session
+       :filter (when (derived-mode-p 'org-mode)
+		 #'ellama--translate-markdown-to-org-filter)
+       :on-done (lambda (res)
+		  (when tl
+		    (ellama-chain res tl)))))))
+
 (defun ellama-chat-done (text)
   "Chat done.
 Will call `ellama-chat-done-callback' on TEXT."
@@ -1045,34 +1403,43 @@ ARGS contains keys for fine control.
   (interactive "sAsk ellama: ")
   (let* ((providers (append
                      `(("default model" . ellama-provider)
-		               ,(if (and ellama-ollama-binary (file-exists-p ellama-ollama-binary))
-			                '("ollama model" . (ellama-get-ollama-local-model))))
+		       ,(if (and ellama-ollama-binary (file-exists-p ellama-ollama-binary))
+			    '("ollama model" . (ellama-get-ollama-local-model))))
                      ellama-providers))
-	     (variants (mapcar #'car providers))
-	     (provider (if current-prefix-arg
-		               (eval (alist-get
-			                  (completing-read "Select model: " variants)
-			                  providers nil nil #'string=))
-		             (or (plist-get args :provider)
-			             ellama-provider)))
-	     (session (if (or create-session
-			              current-prefix-arg
-			              (and (not ellama--current-session)
-			                   (not ellama--current-session-id)))
-		              (ellama-new-session provider prompt)
-		            (or ellama--current-session
-			            (with-current-buffer (ellama-get-session-buffer
-					                          ellama--current-session-id)
-			              ellama--current-session))))
-	     (buffer (ellama-get-session-buffer
-		          (ellama-session-id session)))
-	     (file-name (ellama-session-file session))
-	     (translation-buffer (when ellama-chat-translation-enabled
-			                   (if file-name
-				                   (progn
-				                     (find-file-noselect
-				                      (ellama--get-translation-file-name file-name)))
-				                 (get-buffer-create (ellama-session-id session))))))
+	 (variants (mapcar #'car providers))
+	 (provider (if current-prefix-arg
+		       (eval (alist-get
+			      (completing-read "Select model: " variants)
+			      providers nil nil #'string=))
+		     (or (plist-get args :provider)
+			 ellama-provider)))
+	 (session (if (or create-session
+			  current-prefix-arg
+			  (and provider
+			       (or (plist-get args :provider)
+				   (not (equal provider ellama-provider)))
+			       ellama--current-session-id
+			       (with-current-buffer (ellama-get-session-buffer
+						     ellama--current-session-id)
+				 (not (equal
+				       provider
+				       (ellama-session-provider ellama--current-session)))))
+			  (and (not ellama--current-session)
+			       (not ellama--current-session-id)))
+		      (ellama-new-session provider prompt)
+		    (or ellama--current-session
+			(with-current-buffer (ellama-get-session-buffer
+					      ellama--current-session-id)
+			  ellama--current-session))))
+	 (buffer (ellama-get-session-buffer
+		  (ellama-session-id session)))
+	 (file-name (ellama-session-file session))
+	 (translation-buffer (when ellama-chat-translation-enabled
+			       (if file-name
+				   (progn
+				     (find-file-noselect
+				      (ellama--get-translation-file-name file-name)))
+				 (get-buffer-create (ellama-session-id session))))))
     (if ellama-chat-translation-enabled
 	(ellama--translate-interaction prompt translation-buffer buffer session)
       (display-buffer buffer)
@@ -1080,7 +1447,7 @@ ARGS contains keys for fine control.
 	(save-excursion
 	  (goto-char (point-max))
 	  (insert (ellama-get-nick-prefix-for-mode) " " ellama-user-nick ":\n"
-		  (ellama--format-context session) prompt "\n\n"
+		  (ellama--format-context session) (ellama--fill-long-lines prompt) "\n\n"
 		  (ellama-get-nick-prefix-for-mode) " " ellama-assistant-nick ":\n")
 	  (ellama-stream prompt
 			 :session session
@@ -1188,6 +1555,15 @@ ARGS contains keys for fine control.
 		(buffer-substring-no-properties (point-min) (point-max)))))
     (ellama-instant (format ellama-summarize-prompt-template text))))
 
+;;;###autoload
+(defun ellama-summarize-killring ()
+  "Summarize text from the kill ring."
+  (interactive)
+  (let ((text (current-kill 0)))
+    (if (string-empty-p text)
+        (message "No text in the kill ring to summarize.")
+      (ellama-instant (format ellama-summarize-prompt-template text)))))
+
 ;;;###autoload
 (defun ellama-code-review ()
   "Review code in selected region or current buffer."
@@ -1340,18 +1716,29 @@ buffer."
   (ellama-make-format ellama-make-table-prompt-template))
 
 (defun ellama-summarize-webpage (url)
-  "Summarize webpage fetched from URL."
-  (interactive "sEnter URL you want to summarize: ")
+  "Summarize webpage fetched from URL.
+
+Summarize the URL at point if `thing-at-point' is present, or using
+`shr-url-at-point' if a URL is at point in modes like `eww' or `elfeed',
+otherwise prompt user for URL to summarize."
+  (interactive
+   (list
+    (if-let ((url (or (and (fboundp 'thing-at-point) (thing-at-point 'url))
+                      (shr-url-at-point nil))))
+        url
+      (read-string "Enter URL you want to summarize: "))))
   (let ((buffer-name (url-retrieve-synchronously url t)))
     ;; (display-buffer buffer-name)
     (with-current-buffer buffer-name
       (goto-char (point-min))
-      (search-forward "<!DOCTYPE")
+      (or (search-forward "<!DOCTYPE" nil t)
+          (search-forward "<html" nil))
       (beginning-of-line)
       (kill-region (point-min) (point))
       (shr-insert-document (libxml-parse-html-region (point-min) (point-max)))
       (goto-char (point-min))
-      (search-forward "<!DOCTYPE")
+      (or (search-forward "<!DOCTYPE" nil t)
+          (search-forward "<html" nil))
       (beginning-of-line)
       (kill-region (point) (point-max))
       (ellama-summarize))))
